import json
from abc import ABC, abstractmethod
from typing import Union, List, overload, Dict

from . import api
from .gateway import Requestable
from .interface import LazyLoadable, MessageTypes, ChannelTypes


class Channel(LazyLoadable, Requestable, ABC):
    """
    Interface, represents a channel where messages flowing
    """

    id: str
    type: ChannelTypes

    @abstractmethod
    async def send(self, content: Union[str, List], **kwargs):
        """
        send a msg to the channel
        """
        raise NotImplementedError


class PublicChannel(Channel, ABC):
    name: str
    user_id: str
    guild_id: str
    topic: str
    is_category: int
    parent_id: str
    level: int
    permission_overwrites: list
    permission_users: list
    permission_sync: int

    def __init__(self, **kwargs):
        self.id: str = kwargs.get('id')
        self.name: str = kwargs.get('name')
        self.user_id: str = kwargs.get('user_id')
        self.guild_id: str = kwargs.get('guild_id')
        self.topic: str = kwargs.get('topic')
        self.is_category: int = kwargs.get('is_category')
        self.parent_id: str = kwargs.get('parent_id')
        self.level: int = kwargs.get('level')
        self.type: ChannelTypes = kwargs.get('type') and ChannelTypes(kwargs.get('type'))
        self.permission_overwrites: list = kwargs.get('permission_overwrites')
        self.permission_users: list = kwargs.get('permission_users')
        self.permission_sync: int = kwargs.get('permission_sync')

        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_')


class PublicTextChannel(PublicChannel):
    """
    `Standard Object`

    Text chat channels in guild
    """
    slow_mode: int

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.slow_mode: int = kwargs.get('slow_mode')

    async def load(self):
        pass

    @overload
    async def send(self, content: Union[str, List], **kwargs):
        ...

    async def send(self, content: Union[str, List], *, temp_target_id: str = '', **kwargs):
        """
        send a msg to the channel

        ``temp_target_id`` is available in PublicTextChannel, so ``send()`` is overloaded here
        """
        # if content is card msg, then convert it to plain str
        if isinstance(content, List):
            kwargs['type'] = MessageTypes.CARD.value
            content = json.dumps(content)

        # merge params
        kwargs['target_id'] = self.id
        kwargs['content'] = content
        if temp_target_id:
            kwargs['temp_target_id'] = temp_target_id

        return await self.gate.exec_req(api.Message.create(**kwargs))


class PublicVoiceChannel(PublicChannel):
    """
    Voice chat channel in guild

    a placeholder now for future design/adaption
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def load(self):
        pass

    async def send(self, content: Union[str, List], **kwargs):
        raise TypeError('now there is no PublicVoiceChannel, *hey dude we have a pkg from future*')


class PrivateChannel(Channel):
    """
    Private chat channel
    """

    code: str
    last_read_time: int
    latest_msg_time: int
    unread_count: int
    is_friend: bool
    is_blocked: bool
    is_target_blocked: bool
    target_info: Dict

    def __init__(self, **kwargs):
        self.code: str = kwargs.get('code')
        self.last_read_time: int = kwargs.get('last_read_time')
        self.latest_msg_time: int = kwargs.get('latest_msg_time')
        self.unread_count: int = kwargs.get('unread_count')
        self.is_friend: bool = kwargs.get('is_friend')
        self.is_blocked: bool = kwargs.get('is_blocked')
        self.is_target_blocked: bool = kwargs.get('is_target_blocked')
        self.target_info: Dict = kwargs.get('target_info')

        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_')

    async def load(self):
        pass

    @property
    def target_user_id(self) -> str:
        return self.target_info.get('id') if self.target_info else None

    @property
    def target_user_name(self) -> str:
        return self.target_info.get('username') if self.target_info else None

    @property
    def is_target_user_online(self) -> bool:
        return self.target_info.get('online') if self.target_info else None

    @property
    def target_user_avatar(self) -> str:
        return self.target_info.get('avatar') if self.target_info else None

    async def send(self, content: Union[str, List], **kwargs):
        # if content is card msg, then convert it to plain str
        if isinstance(content, List):
            kwargs['type'] = MessageTypes.CARD.value
            content = json.dumps(content)

        # merge params
        if self.code:
            kwargs['chat_code'] = self.code
        else:
            kwargs['target_id'] = self.target_user_id
        kwargs['content'] = content

        return await self.gate.exec_req(api.DirectMessage.create(**kwargs))



from abc import ABC
from typing import Any, List, Dict, Union

from .interface import MessageTypes, ChannelPrivacyTypes
from .channel import PublicTextChannel, PrivateChannel
from .context import Context
from .gateway import Requestable
from .guild import Guild
from .user import User


class RawMessage(ABC):
    """
    Basic and common features of kinds of messages.

    now support Message kinds:
        1. Message (sent by users, those normal chats such as TEXT/IMG etc.)
        2. Event (sent by system, such as notifications and broadcasts)
    """

    _type: int
    _channel_type: str
    target_id: str
    author_id: str
    content: str
    msg_id: str
    msg_timestamp: int
    nonce: str
    extra: Any

    def __init__(self, **kwargs):
        self._type = kwargs.get('type')
        self._channel_type = kwargs.get('channel_type')
        self.target_id = kwargs.get('target_id')
        self.author_id = kwargs.get('author_id')
        self.content = kwargs.get('content')
        self.msg_id = kwargs.get('msg_id')
        self.msg_timestamp = kwargs.get('msg_timestamp')
        self.nonce = kwargs.get('nonce')
        self.extra = kwargs.get('extra', {})

    @property
    def type(self) -> MessageTypes:
        return MessageTypes(self._type)

    @property
    def channel_type(self) -> ChannelPrivacyTypes:
        return ChannelPrivacyTypes(self._channel_type)


class Message(RawMessage, Requestable, ABC):
    """
    Represents the messages sent by user.

    Because it has source and context, we can interact with its sender and context via it.

    now there are two types of message:
        1. ChannelMessage: sent in a guild channel
        2. PrivateMessage: sent in a private chat
    """
    _ctx: Context

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.gate = kwargs.get('_gate_', None)
        self._author = User(**self.extra['author'], _gate_=self.gate, _lazy_loaded_=True)

    @property
    def author(self) -> User:
        return self._author

    @property
    def ctx(self) -> Context:
        return self._ctx

    async def reply(self, content: Union[str, List] = '', use_quote: bool = True, **kwargs):
        """
        reply to a msg, content can also be a card
        """
        if use_quote:
            kwargs['quote'] = self.msg_id

        return await self.ctx.channel.send(content, **kwargs)


class PublicMessage(Message):
    """
    Messages sent in a `PublicTextChannel`
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        channel = PublicTextChannel(id=self.target_id, name=self.extra['channel_name'], _gate_=self.gate)
        guild = Guild(id=self.extra['guild_id'], _gate_=self.gate)
        self._ctx = Context(channel=channel, guild=guild, _gate_=self.gate)

    @property
    def guild(self) -> Guild:
        return self.ctx.guild

    @property
    def channel(self) -> PublicTextChannel:
        if isinstance(self.ctx.channel, PublicTextChannel):
            return self.ctx.channel
        else:
            raise ValueError('PublicMessage should be placed in PublicTextChannel')

    @property
    def mention(self) -> List[str]:
        return self.extra['mention']

    @property
    def mention_all(self) -> bool:
        return self.extra['mention_all']

    @property
    def mention_roles(self) -> List:  # TODO: check type
        return self.extra['mention_roles']

    @property
    def mention_here(self) -> bool:
        return self.extra['mention_here']

    async def reply(self, content: Union[str, List] = '', use_quote: bool = True, is_temp: bool = False, **kwargs):
        return await super().reply(content, use_quote, temp_target_id=self.author_id if is_temp else '', **kwargs)


class PrivateMessage(Message):
    """
    Messages sent in a `PrivateChannel`
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._channel = PrivateChannel(code=self.extra['code'], target_info=self.extra['author'], _gate_=self.gate)
        self._ctx = Context(channel=self._channel, _gate_=self.gate)

    @property
    def chat_code(self) -> str:
        return self.extra['code']

    @property
    def channel(self) -> PrivateChannel:
        return self._channel


class Event(RawMessage):
    @property
    def event_type(self) -> str:
        return self.extra['type']

    @property
    def body(self) -> Dict:
        return self.extra['body']



import asyncio
import logging
from typing import Dict, Callable, List, Optional, Union, Pattern

from .command import Command
from .lexer import Lexer
from .parser import Parser
from .. import Cert, HTTPRequester, WebhookReceiver, WebsocketReceiver, Gateway, Client
from .. import User, Channel, PublicTextChannel, AsyncRunnable, Message, MessageTypes
from .. import api

log = logging.getLogger(__name__)


class Bot(AsyncRunnable):
    """
    Represents a entity that handles msg/events and interact with users/khl server in manners that programmed.
    """
    client: Client
    _me: Optional[User]
    _cmd_index: Dict[str, Command] = {}

    def __init__(self, *, token: str = '', cert: Cert = None, client: Client = None, gate: Gateway = None,
                 out: HTTPRequester = None, compress: bool = True, port=5000, route='/khl-wh'):
        """
        The most common usage: ``Bot(token='xxxxxx')``

        That's enough.

        :param cert: used to build requester and receiver
        :param client: the bot relies on
        :param gate: the client relies on
        :param out: the gate's component
        :param compress: used to tune the receiver
        :param port: used to tune the WebhookReceiver
        :param route: used to tune the WebhookReceiver
        """
        if not token and not cert:
            raise ValueError('require token or cert')
        self._init_client(cert if cert else Cert(token=token), client, gate, out, compress, port, route)
        self._me = None
        self.client.register(MessageTypes.TEXT, self._make_msg_handler())

    def _init_client(self, cert: Cert, client: Client, gate: Gateway, out: HTTPRequester, compress: bool, port, route):
        """
        construct self.client from args.

        you can init client with kinds of filling ways,
        so there is a priority in the rule: client > gate > out = compress = port = route

        :param cert: used to build requester and receiver
        :param client: the bot relies on
        :param gate: the client relies on
        :param out: the gate's component
        :param compress: used to tune the receiver
        :param port: used to tune the WebhookReceiver
        :param route: used to tune the WebhookReceiver
        :return:
        """
        if client:
            self.client = client
            return
        if gate:
            self.client = Client(gate)
            return

        # client and gate not in args, build them
        _out = out if out else HTTPRequester(cert)
        if cert.type == Cert.Types.WEBSOCKET:
            _in = WebsocketReceiver(cert, compress)
        elif cert.type == Cert.Types.WEBHOOK:
            _in = WebhookReceiver(cert, port=port, route=route, compress=compress)
        else:
            raise ValueError(f'cert type: {cert.type} not supported')

        self.client = Client(Gateway(_out, _in))

    def _make_msg_handler(self) -> Callable:
        """
        construct a function to receive msg from client, and interpret it with _cmd_index
        """

        # use closure
        async def handler(msg: Message):
            for name, cmd in self._cmd_index.items():
                try:
                    args = cmd.prepare(msg)
                except Lexer.LexerException:  # TODO: a more reasonable exception handle in lex and parse
                    continue
                await cmd.execute(msg, *args)

        return handler

    def add_command(self, cmd: Command) -> Command:
        """
        register the cmd on current Bot

        :param cmd: the Command going to be registered
        :return: the cmd
        """
        if cmd.name in self._cmd_index:
            raise ValueError(f'cmd: {cmd.name} already exists')
        self._cmd_index[cmd.name] = cmd
        log.debug(f'cmd: {cmd.name} added')
        return cmd

    def command(self, name: str = '', *, help: str = '', desc: str = '',
                aliases: List[str] = (), prefixes: List[str] = ('/',), regex: Union[str, Pattern] = '',
                lexer: Lexer = None, parser: Parser = None):
        """
        decorator, wrap a function in Command and register it on current Bot

        :param name: the name of this Command, also used to trigger command in DefaultLexer
        :param aliases: (DefaultLexer only) you can also trigger the command with aliases
        :param prefixes: (DefaultLexer only) command prefix, default use '/'
        :param regex: (RELexer only) pattern for the command
        :param help: detailed manual
        :param desc: short introduction
        :param lexer: (Advanced) explicitly set the lexer
        :param parser: (Advanced) explicitly set the parser
        :return: wrapped Command
        """
        args = {'help': help, 'desc': desc,
                'aliases': aliases, 'prefixes': prefixes, 'regex': regex,
                'lexer': lexer, 'parser': parser}

        # use lambda cuz i do not wanna declare decorator() explicitly to take 3 blank lines
        # did not init Lexer in advance cuz it needs func.__name__
        # this is redundant stuff in constructor, there should be a better way
        return lambda func: self.add_command(Command.command(name, **args)(func))

    async def fetch_me(self, force_update: bool = False) -> User:
        if force_update or not self._me or not self._me.is_loaded():
            self._me = User(**(await self.client.gate.exec_req(api.User.me())), _lazy_loaded_=True)
        return self._me

    @property
    def me(self) -> User:
        """
        get bot it self's data

        RECOMMEND: use ``await fetch_me()``

        CAUTION: please call ``await fetch_me()`` first to load data from khl server

        designed as 'empty-then-fetch' will break the rule 'net-related is async'

        :return: the bot's underlying User
        """
        if self._me and self._me.is_loaded():
            return self._me
        raise ValueError('not loaded, please call `await fetch_me()` first')

    @staticmethod
    async def send(target: Channel, content: Union[str, List], *, temp_target_id: str = '', **kwargs):
        """
        send a msg to a channel

        ``temp_target_id`` is only available in ChannelPrivacyTypes.GROUP
        """
        if isinstance(target, PublicTextChannel):
            kwargs['temp_target_id'] = temp_target_id

        return await target.send(content, **kwargs)

    def run(self):
        try:
            if not self.loop:
                self.loop = asyncio.get_event_loop()
            self.loop.run_until_complete(self.client.run())
        except KeyboardInterrupt:
            pass
        except Exception as e:
            log.error(e)
        log.info('see you next time')



class Role:
    """
    `Standard Object`

    represent the component that used in permission control and user identify
    """
    role_id: int
    name: str
    color: int
    position: int
    hoist: int
    mentionable: int
    permissions: int

    def __init__(self, **kwargs):
        self.role_id: int = kwargs.get("role_id", 0)
        self.name: str = kwargs.get("name", "")
        self.color: int = kwargs.get("color", 0)
        self.position: int = kwargs.get("position", 0)
        self.hoist: int = kwargs.get("hoist", 0)
        self.mentionable: int = kwargs.get("mentionable", 0)
        self.permissions: int = kwargs.get("permissions", 0)



import asyncio
import inspect
import logging
from typing import Dict, List, Callable

from .gateway import Gateway
from .gateway import Requestable
from .interface import AsyncRunnable, MessageTypes
from .message import RawMessage, Event, PublicMessage, PrivateMessage

log = logging.getLogger(__name__)


class Client(Requestable, AsyncRunnable):
    """
    The bridge between khl.py internal components and khl server.

    Translates network package into khl.py concepts/object for internal to use.

    reminder: Client.loop only used to run handle_event() and registered handlers.
    """
    _handler_map: Dict[MessageTypes, List[Callable]]

    def __init__(self, gate: Gateway):
        self.gate = gate

        self._handler_map = {}
        self._pkg_queue = asyncio.Queue()

    def register(self, type: MessageTypes, handler: Callable):
        if not asyncio.iscoroutinefunction(handler):
            raise TypeError('handler must be a coroutine.')

        params = list(inspect.signature(handler).parameters.items())
        if len(params) != 1 or not issubclass(params[0][1].annotation, RawMessage):
            raise TypeError('handler must have one and only one param, and the param inherits RawMessage')

        if type not in self._handler_map:
            self._handler_map[type] = []
        self._handler_map[type].append(handler)

    async def handle_pkg(self):
        """
        Pop `pkg` from `event_queue`,
        spawn `msg` according to `pkg`,
        pass `msg` to corresponding handlers defined in `_handler_map`
        """
        while True:
            pkg: Dict = await self._pkg_queue.get()
            log.debug(f'upcoming pkg: {pkg}')

            msg: RawMessage
            if pkg['type'] == MessageTypes.SYS.value:
                msg = Event(**pkg)
            else:
                if pkg['channel_type'] == 'GROUP':
                    msg = PublicMessage(**pkg, _gate_=self.gate)
                elif pkg['channel_type'] == 'PERSON':
                    msg = PrivateMessage(**pkg, _gate_=self.gate)
                else:
                    log.error(f'can not spawn msg from pkg: {pkg}')
                    self._pkg_queue.task_done()
                    continue

            # dispatch msg
            if msg.type in self._handler_map and self._handler_map[msg.type]:
                for handler in self._handler_map[msg.type]:
                    asyncio.ensure_future(handler(msg), loop=self.loop)

            self._pkg_queue.task_done()

    async def run(self):
        asyncio.ensure_future(self.handle_pkg(), loop=self.loop)
        await self.gate.run(self._pkg_queue)



import logging
import re
import shlex
from abc import ABC, abstractmethod
from typing import List, Set, Union, Pattern

from ..message import Message

log = logging.getLogger(__name__)


class Lexer(ABC):
    """
    deal with msg, convert msg.content -> a list of tokens(like tokens in compiler)
    """

    @abstractmethod
    def lex(self, msg: Message) -> List[str]:
        """
        read msg, and get a list of tokens as List[str] for following parser.parse()

        :param msg: which will be torn down into List[str]
        :return: Tuple(is the msg match the lexer, thrived tokens)
        """
        raise NotImplementedError

    class LexerException(Exception):
        def __init__(self, msg: Message):
            self.msg = msg


class DefaultLexer(Lexer):
    """
    lex with python standard module ``shlex``
    """
    prefixes: Set[str]
    triggers: Set[str]

    def __init__(self, prefixes: Set[str], triggers: Set[str]):
        self.prefixes = prefixes
        self.triggers = triggers

    def lex(self, msg: Message) -> List[str]:
        """
        checks if msg starts with any prefix, shlex.split() if hit, then check if starts with any trigger

        :param msg: which will be torn down into List[str]
        :return: Tuple(is the msg match the lexer, thrived tokens)
        :raises:    ShlexLexer.MalformedContent if shlex.split() failed
                    ShlexLexer.NoMatchedPrefix  literally
                    ShlexLexer.NoMatchedTrigger literally
        """
        for prefix in [p for p in self.prefixes if msg.content.startswith(p)]:  # get matched prefix
            # tear down
            try:
                arg_list = shlex.split(msg.content)
            except Exception:
                raise DefaultLexer.MalformedContent(msg)

            # check if trigger exists
            if arg_list[0][len(prefix):] in self.triggers:
                # exists
                return arg_list[1:]  # arg_list[0] is trigger
            else:
                # not exists
                raise DefaultLexer.NoMatchedTrigger(msg)

        # no matched prefix(did not enter the for-loop)
        raise DefaultLexer.NoMatchedPrefix(msg)

    class MalformedContent(Lexer.LexerException):
        pass

    class NoMatchedPrefix(Lexer.LexerException):
        pass

    class NoMatchedTrigger(Lexer.LexerException):
        pass


class RELexer(Lexer):
    """
    lex with python built-in module ``re``
    """
    pattern: Pattern

    def __init__(self, regex: Union[str, Pattern]):
        if isinstance(regex, str):
            self.pattern = re.compile(regex)
        else:
            self.pattern = regex

    def lex(self, msg: Message) -> List[str]:
        m = self.pattern.fullmatch(msg.content)
        if not m:
            raise RELexer.NotMatched(msg)
        if m.groups():
            return [m.group(i) for i in range(1, len(m.groups())) if m.start(i) < len(msg.content)]

    class NotMatched(Lexer.LexerException):
        pass



import asyncio
from abc import ABC
from typing import Union, List

from .api import _Req
from .receiver import Receiver
from .requester import HTTPRequester


class Gateway:
    """
    Component which deals with network connection and package send/receive

    reminder: this is not AsyncRunnable cuz gateway dose not have its own tasks, only pass loop to _in/_out
    """
    requester: HTTPRequester
    receiver: Receiver

    def __init__(self, requester: HTTPRequester, receiver: Receiver):
        self.requester = requester
        self.receiver = receiver

    async def request(self, method: str, route: str, **params) -> Union[dict, list]:
        """
        execute raw request, this is just a wrapper for convenience
        """
        return await self.requester.request(method, route, **params)

    async def exec_req(self, r: _Req):
        """
        execute request, this is just a wrapper for convenience
        """
        return await self.requester.exec_req(r)

    async def exec_pagination_req(self, r: _Req, page: int = 1, page_size: int = 50, sort: str = '') -> List:
        return await self.requester.exec_pagination_req(r, page, page_size, sort)

    async def run(self, in_queue: asyncio.Queue):
        await self.receiver.run(in_queue)


class Requestable(ABC):
    """
    Classes that can use a `Gateway` to communicate with khl server.

    For example:
        `Message`: can use msg.reply() to send a reply to khl

        `Guild`: guild.get_roles() to fetch role list from khl
    """
    gate: Gateway



import functools
import inspect
import logging
import re
from typing import Dict, Callable

log = logging.getLogger(__name__)


class _Req:
    def __init__(self, method: str, route: str, params: Dict):
        self.method = method
        self.route = route
        self.params = params


def req(method: str, **http_fields):
    def _method(func: Callable):
        @functools.wraps(func)
        def req_maker(*args, **kwargs) -> _Req:
            route = re.sub(r'(?<!^)(?=[A-Z])', '-', func.__qualname__).lower().replace('.', '/')

            # dump args into kwargs
            param_names = list(inspect.signature(func).parameters.keys())
            for i in range(len(args)):
                kwargs[param_names[i]] = args[i]

            # merge http_fields with kwargs
            params = {'json' if method == 'POST' else 'params': kwargs}
            params.update(http_fields)

            return _Req(method, route, params)

        return req_maker

    return _method


class Guild:

    @staticmethod
    @req('GET')
    def list(
    ):
        ...

    @staticmethod
    @req('GET')
    def view(
            guild_id
    ):
        ...

    @staticmethod
    @req('GET')
    def userList(
            guild_id,
            channel_id,
            search,
            role_id,
            mobile_verified,
            active_time,
            joined_at,
            page,
            page_size,
    ):
        ...

    @staticmethod
    @req('POST')
    def nickname(
            guild_id,
            nickname,
            user_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def leave(
            guild_id
    ):
        ...

    @staticmethod
    @req('POST')
    def kickout(
            guild_id,
            target_id
    ):
        ...


class GuildMute:

    @staticmethod
    @req('GET')
    def list(
            guild_id,
            return_type
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            guild_id,
            user_id,
            type,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            guild_id,
            user_id,
            type,
    ):
        ...


class Channel:

    @staticmethod
    @req('GET')
    def list(
            guild_id
    ):
        ...

    @staticmethod
    @req('GET')
    def view(
            target_id
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            guild_id,
            parent_id,
            name,
            type,
            limit_amount,
            voice_quality,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            channel_id
    ):
        ...

    @staticmethod
    @req('POST')
    def moveUser(
            target_id,
            user_ids
    ):
        ...


class ChannelRole:

    @staticmethod
    @req('GET')
    def index(
            channel_id
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            channel_id,
            type,
            value,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
            channel_id,
            type,
            value,
            allow,
            deny,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            channel_id,
            type,
            value,
    ):
        ...


class Message:

    @staticmethod
    @req('GET')
    def list(
            target_id,
            msg_id,
            pin,
            flag,
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            type,
            target_id,
            content,
            quote,
            nonce,
            temp_target_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
            msg_id,
            content,
            quote,
            temp_target_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            msg_id,
    ):
        ...

    @staticmethod
    @req('GET')
    def reactionList(
            msg_id,
            emoji
    ):
        ...

    @staticmethod
    @req('POST')
    def addReaction(
            msg_id,
            emoji
    ):
        ...

    @staticmethod
    @req('POST')
    def deleteReaction(
            msg_id,
            emoji,
            user_id
    ):
        ...

    @staticmethod
    @req('POST')
    def addReaction(
            msg_id,
            emoji
    ):
        ...

    @staticmethod
    @req('POST')
    def addReaction(
            msg_id,
            emoji
    ):
        ...


class UserChat:

    @staticmethod
    @req('GET')
    def list(
    ):
        ...

    @staticmethod
    @req('GET')
    def view(
            chat_code
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            target_id
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            chat_code
    ):
        ...


class DirectMessage:

    @staticmethod
    @req('GET')
    def list(
            chat_code,
            target_id,
            msg_id,
            flag,
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            type,
            target_id,
            chat_code,
            content,
            quote,
            nonce,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
            msg_id,
            content,
            quote,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            msg_id
    ):
        ...

    @staticmethod
    @req('GET')
    def reactionList(
            msg_id,
            emoji
    ):
        ...

    @staticmethod
    @req('POST')
    def addReaction(
            msg_id,
            emoji
    ):
        ...

    @staticmethod
    @req('POST')
    def deleteReaction(
            msg_id,
            emoji
    ):
        ...


class Gateway:

    @staticmethod
    @req('GET')
    def index(
            compress
    ):
        ...


class User:

    @staticmethod
    @req('GET')
    def me(
    ):
        ...

    @staticmethod
    @req('GET')
    def view(
            user_id,
            guild_id,
    ):
        ...


class Asset:

    @staticmethod
    @req('POST')
    def create(
            file
    ):
        ...


class GuildRole:

    @staticmethod
    @req('GET')
    def list(
            guild_id
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            name,
            guild_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
            guild_id,
            role_id,
            hoist,
            mentionable,
            permissions,
            color,
            name,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            guild_id,
            role_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def grant(
            guild_id,
            user_id,
            role_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def revoke(
            guild_id,
            user_id,
            role_id,
    ):
        ...


class Intimacy:

    @staticmethod
    @req('GET')
    def index(
            user_id
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
            user_id,
            score,
            social_info,
            img_id,
    ):
        ...


class GuildEmoji:

    @staticmethod
    @req('GET')
    def list(
            guild_id
    ):
        ...

    @staticmethod
    @req('POST', headers={'Content-Type': 'multipart/form-data'})
    def create(
            name,
            guild_id,
            emoji,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
            name,
            id,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            id
    ):
        ...


class Invite:

    @staticmethod
    @req('GET')
    def list(
            guild_id,
            channel_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            guild_id,
            channel_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            guild_id,
            channel_id,
            url_code,
    ):
        ...



import asyncio
import logging
from typing import Callable, Coroutine, List, Any, Union, Pattern

from .lexer import Lexer, RELexer, DefaultLexer
from .parser import Parser
from ..message import Message

log = logging.getLogger(__name__)


class Command:
    """
    Receive a `Message`, parse it with `lex` and `parser`, execute `handler` if successfully parsed else discard it
    """
    name: str
    handler: Callable[..., Coroutine]
    help: str
    desc: str

    lexer: Lexer
    parser: Parser

    def __init__(self, name: str, handler: Callable, help: str, desc: str, lexer: Lexer, parser: Parser):
        if not asyncio.iscoroutinefunction(handler):
            raise TypeError('handler must be a coroutine.')
        self.handler = handler

        self.name = name or handler.__name__
        if not isinstance(self.name, str):
            raise TypeError('Name of a command must be a string.')

        self.help = help
        self.desc = desc

        self.lexer = lexer
        self.parser = parser

    def _lex(self, msg: Message) -> List[str]:
        return self.lexer.lex(msg)

    def _parse(self, tokens: List[str], handler: Callable) -> List[Any]:
        return self.parser.parse(tokens, handler)

    def prepare(self, msg: Message) -> List[Any]:
        """
        parse msg, prepare arg list from the msg

        :param msg: the msg going to be lexed and parsed
        :return:
        """
        tokens = self._lex(msg)
        args = self._parse(tokens, self.handler)
        return args

    async def execute(self, msg: Message, *args):
        """
        pass msg and args from prepare() to handler()

        :param msg:
        :param args: the returned list of prepare()
        :return:
        """
        log.info(f'command {self.name} was triggered by msg: {msg.content}')
        try:
            await self.handler(msg, *args)
        except Exception:
            raise Command.ExecuteException(self.handler)

    def do(self, msg: Message):
        """
        wrapper for prepare() and execute()

        :param msg: what wanna be handled
        """
        self.execute(msg, self.prepare(msg))

    @staticmethod
    def command(name: str = '', *, help: str = '', desc: str = '',
                aliases: List[str] = (), prefixes: List[str] = ('/',), regex: Union[str, Pattern] = '',
                lexer: Lexer = None, parser: Parser = None):
        """
        decorator, to wrap a func into a Command

        :param name: the name of this Command, also used to trigger command in DefaultLexer
        :param aliases: (DefaultLexer only) you can also trigger the command with aliases
        :param prefixes: (DefaultLexer only) command prefix, default use '/'
        :param regex: (RELexer only) pattern for the command
        :param help: detailed manual
        :param desc: short introduction
        :param lexer: (Advanced) explicitly set the lexer
        :param parser: (Advanced) explicitly set the parser
        :return: wrapped Command
        """
        if not lexer and regex:
            lexer = RELexer(regex)

        # use lambda cuz i do not wanna declare decorator() explicitly to take 3 blank lines
        # did not init Lexer in advance cuz it needs func.__name__
        # this is redundant stuff in constructor, there should be a better way
        return lambda func: Command(name, func, help, desc,
                                    lexer or DefaultLexer(set(prefixes), set([name or func.__name__] + list(aliases))),
                                    parser or Parser())

    class ExecuteException(Exception):
        def __init__(self, func):
            self.func = func



import asyncio
import copy
import inspect
import logging
from typing import Dict, Any, Callable, List, Tuple

log = logging.getLogger(__name__)


class Parser:
    """
    deal with a list of tokens made from Lexer, convert their type to match the command.handler
    """
    _parse_funcs: Dict[Any, Callable] = {
        str: lambda token: token,
        int: lambda token: int(token),
        float: lambda token: float(token)
        # TODO: tag -> User/Channel/Role...
    }

    def __init__(self):
        self._parse_funcs = copy.copy(Parser._parse_funcs)

    def parse(self, tokens: List[str], handler: Callable) -> List[Any]:
        """
        parse tokens into args that types corresponding to handler's requirement

        :param tokens: output of Lexer.lex()
        :param handler: parse target
        :return: List of args
        :raise: Parser.ArgListLenNotMatch
        """
        s = inspect.signature(handler)
        params = list(s.parameters.items())[1:]  # the first param is `msg: Message`

        # check
        if len(tokens) > len(params):
            raise Parser.TooMuchArgs(len(params), len(tokens), handler)

        # parse
        ret = []
        for i in range(len(tokens)):
            t = params[i][1].annotation  # arg type

            # no type hint for t
            if t == inspect.Parameter.empty:
                ret.append(tokens[i])
                continue

            if t not in self._parse_funcs:
                raise Parser.ParseFuncNotExists(params[i], handler)
            try:
                ret.append(self._parse_funcs[t](tokens[i]))
            except Exception as e:
                raise Parser.ParseFuncException(e)
        return ret

    def register(self, func):  # TODO: global register
        """
        decorator, register the func into object restricted _parse_funcs()

        checks if parse func for that type exists, and insert if not
        :param func: parse func
        """
        s = inspect.signature(func)

        # check: 1. not coroutine, 2. len matches
        if asyncio.iscoroutinefunction(func):
            raise TypeError('parse function should not be async')
        if len(s.parameters) != 1 or list(s.parameters.values())[0].annotation != str:
            raise TypeError('parse function should own only one param, and the param type is str')

        # insert, remember this is a replace
        self._parse_funcs[s.return_annotation] = func
        return func

    class ParserException(Exception):
        pass

    class TooMuchArgs(ParserException):
        def __init__(self, expected: int, exact: int, func: Callable):
            self.expected = expected
            self.exact = exact
            self.func = func

    class ParseFuncNotExists(ParserException):
        def __init__(self, expected: Tuple[str, inspect.Parameter], func: Callable):
            self.expected = expected
            self.func = func

    class ParseFuncException(ParserException):
        def __init__(self, err: Exception):
            self.err = err



from typing import List, Union

from . import api
from .channel import PrivateChannel
from .gateway import Requestable
from .interface import LazyLoadable
from .role import Role


class User(LazyLoadable, Requestable):
    """
    `Standard Object`

    represent a entity that interact with khl server
    """
    id: str
    username: str
    nickname: str
    identify_num: str
    online: bool
    bot: bool
    status: int
    avatar: str
    vip_avatar: str
    mobile_verified: bool
    roles: List[Role]

    _loaded: bool
    _channel: PrivateChannel

    def __init__(self, **kwargs):
        self.id = kwargs.get('id', '')
        self.username = kwargs.get('username', '')
        self.nickname = kwargs.get('nickname', '')
        self.identify_num = kwargs.get('identify_num', '')
        self.online = kwargs.get('online', False)
        self.bot = kwargs.get('bot', False)
        self.status = kwargs.get('status', 0)
        self.avatar = kwargs.get('avatar', '')
        self.vip_avatar = kwargs.get('vip_avatar', '')
        self.mobile_verified = kwargs.get('mobile_verified', False)
        self.roles = kwargs.get('roles', [])

        self._loaded = kwargs.get('_lazy_loaded_', False)
        self._channel = kwargs.get('_channel_')
        self.gate = kwargs.get('_gate_', None)

    async def load(self):
        pass

    async def send(self, content: Union[str, List], **kwargs):
        """
        send a msg to a channel

        ``temp_target_id`` is only available in ChannelPrivacyTypes.GROUP
        """
        if not self._channel:
            self._channel = PrivateChannel(**(await self.gate.exec_req(api.UserChat.create(target_id=self.id))),
                                           _lazy_loaded_=True, _gate_=self.gate)

        return await self._channel.send(content, **kwargs)



import asyncio
import json
import logging
import time
import zlib
from abc import ABC, abstractmethod
from typing import Dict

from aiohttp import ClientWebSocketResponse, ClientSession, web, WSMessage

from .cert import Cert
from .interface import AsyncRunnable

log = logging.getLogger(__name__)

API = f'https://www.kaiheila.cn/api/v3'


class Receiver(AsyncRunnable, ABC):
    @abstractmethod
    def __init__(self):
        raise NotImplementedError

    @property
    def type(self) -> str:
        raise NotImplementedError

    @abstractmethod
    async def run(self, pkg_queue: asyncio.Queue):
        raise NotImplementedError


class WebsocketReceiver(Receiver):

    def __init__(self, cert: Cert, compress: bool):
        self._cert = cert
        self.compress = compress

        self._NEWEST_SN = 0
        self._RAW_GATEWAY = ''

    @property
    def type(self) -> str:
        return 'websocket'

    async def heartbeat(self, ws_conn: ClientWebSocketResponse):
        while True:
            await asyncio.sleep(26)
            await ws_conn.send_json({'s': 2, 'sn': self._NEWEST_SN})

    def __raw_to_pkg(self, data: bytes) -> dict:
        data = zlib.decompress(data) if self.compress else data
        data = json.loads(str(data, encoding='utf-8'))
        return data

    async def run(self, pkg_queue: asyncio.Queue):
        async with ClientSession(loop=self.loop) as cs:
            headers = {
                'Authorization': f'Bot {self._cert.token}',
                'Content-type': 'application/json'
            }
            params = {'compress': 1 if self.compress else 0}
            async with cs.get(f"{API}/gateway/index",
                              headers=headers,
                              params=params) as res:
                res_json = await res.json()
                if res_json['code'] != 0:
                    log.error(f'error getting gateway: {res_json}')
                    return

                self._RAW_GATEWAY = res_json['data']['url']

            async with cs.ws_connect(self._RAW_GATEWAY) as ws_conn:
                asyncio.ensure_future(self.heartbeat(ws_conn), loop=self.loop)

                log.info('[ init ] launched')

                async for raw in ws_conn:
                    try:
                        raw: WSMessage
                        pkg: Dict = self.__raw_to_pkg(raw.data)
                        log.debug(f'upcoming raw: {pkg}')
                    except Exception as e:
                        log.exception(e)
                        continue
                    if pkg['s'] == 0:
                        log.debug(f'upcoming pkg: {pkg}')
                        self._NEWEST_SN = pkg['sn']
                        event = pkg['d']
                        await pkg_queue.put(event)


class WebhookReceiver(Receiver):
    def __init__(self, cert: Cert, *, port: int, route: str, compress: bool):
        self._cert = cert
        self.port = port
        self.route = route
        self.cs = ClientSession()
        self.app = web.Application()
        self.compress = compress
        self.sn_dup_map = {}

    @property
    def type(self) -> str:
        return 'webhook'

    def __raw_to_pkg(self, data: bytes) -> dict:
        data = zlib.decompress(data) if self.compress else data
        data = json.loads(str(data, encoding='utf-8'))
        return ('encrypt' in data.keys()) and json.loads(self._cert.decrypt(data['encrypt'])) or data

    def __is_pkg_dup(self, req: dict) -> bool:
        # TODO: need a recycle count down ring
        sn = req.get('sn', None)
        if not sn:
            return False
        current = time.time()
        if sn in self.sn_dup_map.keys():
            if current - self.sn_dup_map[sn] <= 600:
                # 600 sec timed out
                return True
        self.sn_dup_map[sn] = current
        return False

    async def run(self, pkg_queue: asyncio.Queue):
        async def on_recv(request: web.Request):
            try:
                pkg: Dict = self.__raw_to_pkg(await request.read())
            except Exception as e:
                log.exception(e)
                return web.Response()

            if not pkg or pkg['d']['verify_token'] != self._cert.verify_token:
                # empty pkg or verify_token check failed
                return web.Response()

            if self.__is_pkg_dup(pkg):
                # dupe pkg
                return web.Response()

            if pkg['s'] == 0:
                log.debug(f'upcoming pkg: {pkg}')
                pkg = pkg['d']
                if pkg['type'] == 255 and pkg['channel_type'] == 'WEBHOOK_CHALLENGE':
                    return web.json_response({'challenge': pkg['challenge']})
                await pkg_queue.put(pkg)

            return web.Response()

        self.app.router.add_post(self.route, on_recv)

        async def on_shutdown(app):
            await self.cs.close()

        self.app.on_shutdown.append(on_shutdown)

        runner = web.AppRunner(self.app, access_log_class=None)
        await runner.setup()  # runner use its own loop, can not be set
        site = web.TCPSite(runner, '0.0.0.0', self.port)

        log.info('[ init ] launched')

        await site.start()

        while True:
            await asyncio.sleep(3600)  # sleep forever



r"""Some Interfaces"""
import asyncio
from abc import ABC, abstractmethod
from enum import IntEnum, Enum


class AsyncRunnable(ABC):
    """
    Classes that has async work to run
    """
    _loop: asyncio.AbstractEventLoop = None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @loop.setter
    def loop(self, new_loop: asyncio.AbstractEventLoop):
        self._loop = new_loop


class LazyLoadable(ABC):
    """
    Classes that can be initialized before loaded full data from khl server.
    These classes objects usually are constructed by khl.py internal calls.

    For example:
        `Channel`: we usually construct a channel with a message for convenient,
        while we only know the channel's id, so this channel is not `loaded`, until call the `load()`
    """
    _loaded: bool

    @abstractmethod
    async def load(self):
        """
        Load full data from khl server

        :return: empty
        """
        raise NotImplementedError

    def is_loaded(self) -> bool:
        """
        Check if loaded

        :return: bool
        """
        return self._loaded


class MessageTypes(IntEnum):
    """
    types of message, type==SYS will be interpreted as `Event`, others are `Message`
    """
    TEXT = 1
    IMG = 2
    VIDEO = 3
    FILE = 4
    AUDIO = 8
    KMD = 9
    CARD = 10
    SYS = 255


class ChannelTypes(IntEnum):
    """
    types of channel
    """
    CATEGORY = 0
    TEXT = 1
    VOICE = 2


class ChannelPrivacyTypes(Enum):
    """
    channel's privacy level
    """
    GROUP = 'GROUP'
    PERSON = 'PERSON'



import asyncio
import logging
from typing import Union, List

from aiohttp import ClientSession

from .api import _Req
from .cert import Cert
from .interface import AsyncRunnable

log = logging.getLogger(__name__)

API = f'https://www.kaiheila.cn/api/v3'


class HTTPRequester(AsyncRunnable):
    def __init__(self, cert: Cert):
        self._cert = cert
        self._cs: ClientSession = ClientSession(loop=self.loop)

    def __del__(self):
        asyncio.get_event_loop().run_until_complete(self._cs.close())

    async def request(self, method: str, route: str, **params) -> Union[dict, list]:
        headers = params.pop('headers', {})
        headers['Authorization'] = f'Bot {self._cert.token}'
        params['headers'] = headers

        async with self._cs.request(method, f'{API}/{route}', **params) as res:
            log.debug(f'req: {method} {route}: {params}')
            rsp = await res.json()
            if rsp['code'] != 0:
                raise HTTPRequester.APIRequestFailed(method, route, params, rsp['code'], rsp['message'])
            else:
                log.debug(f'req done: {rsp}')
            return rsp['data']

    async def exec_req(self, r: _Req):
        return await self.request(r.method, r.route, **r.params)

    async def exec_pagination_req(self, r: _Req, page: int = 1, page_size: int = 50, sort: str = '') -> List:
        """
        exec pagination requests, iter from ``page`` to the end

        1. get a req, inject the params
        2. req and receive the result
        3. unwrap the result, append result, fresh pagination params
        """
        ret = []
        page_total = page + 1
        while page < page_total:
            r.params['params']['page'] = page
            r.params['params']['page_size'] = page_size
            if sort:
                r.params['params']['sort'] = sort

            p = await self.exec_req(r)

            ret.extend(p['items'])
            page = p['meta']['page']
            page_total = p['meta']['page_total']
            page_size = p['meta']['page_size']

        return ret

    class APIRequestFailed(Exception):
        def __init__(self, method, route, params, err_code, err_message):
            self.method = method
            self.route = route
            self.params = params
            self.err_code = err_code
            self.err_message = err_message



from typing import List, Dict

from . import api
from .channel import Channel, PublicTextChannel, PublicVoiceChannel
from .gateway import Requestable
from .interface import LazyLoadable, ChannelTypes
from .role import Role


class Guild(LazyLoadable, Requestable):
    """
    `Standard Object`

    represent a server where users gathered in and contains channels
    """
    id: str
    name: str
    topic: str
    master_id: str
    icon: str
    notify_type: int
    region: str
    enable_open: bool
    open_id: str
    default_channel_id: str
    welcome_channel_id: str
    _roles: List[Role]
    _channel_categories: List[Dict]
    _channels: List[Channel]

    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name', '')
        self.topic = kwargs.get('topic', '')
        self.master_id = kwargs.get('master_id', '')
        self.icon = kwargs.get('icon', '')
        self.notify_type = kwargs.get('notify_type', 0)
        self.region = kwargs.get('region', '')
        self.enable_open = kwargs.get('enable_open', False)
        self.open_id = kwargs.get('open_id', '')
        self.default_channel_id = kwargs.get('default_channel_id', '')
        self.welcome_channel_id = kwargs.get('welcome_channel_id', '')
        self._roles = kwargs.get('roles', None)
        self._channel_categories = []
        self._channels = kwargs.get('channels', None)

        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_', None)

    async def load(self):
        pass

    async def fetch_channel_list(self, force_update: bool = True) -> List[Channel]:
        if force_update or self._channels is None:
            raw_list = await self.gate.exec_pagination_req(api.Channel.list(guild_id=self.id))
            channel_list: List[Channel] = []
            for i in raw_list:
                if i['type'] == ChannelTypes.CATEGORY:
                    self._channel_categories.append(i)
                elif i['type'] == ChannelTypes.TEXT:
                    channel_list.append(PublicTextChannel(**i))
                elif i['type'] == ChannelTypes.VOICE:
                    channel_list.append(PublicVoiceChannel(**i))
            self._channels = channel_list
        return self._channels

    @property
    def channels(self) -> List[Channel]:
        """
        get guild's channel list

        RECOMMEND: use ``await fetch_channel_list()``

        CAUTION: please call ``await fetch_me()`` first to load data from khl server

        designed as 'empty-then-fetch' will break the rule 'net-related is async'
        """
        if self._channels is not None:
            return self._channels
        raise ValueError('not loaded, please call `await fetch_channel_list()` first')

    async def fetch_roles(self, force_update: bool = True) -> List[Role]:
        if force_update or self._roles is None:
            raw_list = await self.gate.exec_pagination_req(api.GuildRole.list(guild_id=self.id))
            self._roles = [Role(**i) for i in raw_list]
        return self._roles

    async def create_role(self, role_name: str) -> Role:
        return Role(**(await self.gate.exec_req(api.GuildRole.create(guild_id=self.id, name=role_name))))

    async def update_role(self, new_role: Role) -> Role:
        return Role(**(await self.gate.exec_req(api.GuildRole.update(guild_id=self.id, **vars(new_role)))))

    async def delete_role(self, role_id: int):
        return await self.gate.exec_req(api.GuildRole.delete(guild_id=self.id, role_id=role_id))
